r"""Configure the Sphinx documentation builder.

https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

from zathura_language_server import __version__ as version  # type: ignore
from zathura_language_server._metainfo import (  # type: ignore
    author,
    copyright,
    project,
)

__all__ = ["version", "author", "copyright", "project"]

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

# -- Project information -----------------------------------------------------
language = "en"
locale_dirs = ["locale"]
gettext_compact = False

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.githubpages",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "myst_parser",
    "sphinxcontrib.autofile",
    "sphinxcontrib.requirements_txt",
]

myst_heading_anchors = 3
myst_title_to_header = True
todo_include_todos = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]



r"""Zathurarc
=============
"""

from typing import Any

from lsp_tree_sitter.misc import get_md_tokens

from .._metainfo import SOURCE, project


def init_schema() -> dict[str, Any]:
    r"""Init schema.

    :rtype: dict[str, Any]
    """
    filetype = "zathurarc"
    schemas = {
        filetype: {
            "$id": (
                f"{SOURCE}/blob/main/"
                f"src/termux_language_server/assets/json/{filetype}.json"
            ),
            "$schema": "http://json-schema.org/draft-07/schema#",
            "$comment": (
                "Don't edit this file directly! It is generated by "
                f"`{project} --generate-schema={filetype}`."
            ),
            "type": "object",
            "additionalProperties": False,
            "properties": {},
        }
    }
    tokens = get_md_tokens("zathurarc")
    indices = []
    end_index = len(tokens)
    keys = []
    functions = []
    arguments = []
    for i, token in enumerate(tokens):
        if token.content == "OPTIONS":
            end_index = i
            break
        if token.tag == "h2" and token.type == "heading_open":
            indices += [i + 1]
    for i, index in enumerate(indices):
        keyword, _, description = tokens[index].content.partition(" - ")
        schemas[filetype]["properties"][keyword] = {
            "type": "object",
        }
        index2 = end_index if len(indices) - 1 == i else indices[i + 1]
        data = []
        for token in tokens[index + 1 : index2]:
            if len(data) == 2:
                break
            if token.content == "" or token.content.startswith("<!--"):
                continue
            data += [token.content]
        if keyword == "map":
            start = False
            for k, token in enumerate(tokens[index + 1 : index2], index + 1):
                if token.content.replace("*", "") in {
                    "Special keys",
                    "Mouse buttons",
                }:
                    keys += [
                        "<" + line.lstrip("> ").partition(" ")[0] + ">"
                        for line in tokens[k + 5].content.splitlines()
                    ][2:]
                if token.content.replace("*", "") == "Shortcut functions":
                    functions = [
                        line.strip("- :").replace("*", " ").strip(" ")
                        for line in tokens[k + 5].content.splitlines()
                        if line.startswith("-   ")
                    ]
                if token.content.replace("*", "") == "Pass arguments":
                    start = True
                if (
                    start
                    and token.content.islower()
                    and token.type == "inline"
                ):
                    arguments += [token.content]
        schemas[filetype]["properties"][keyword][
            "description"
        ] = f"""# {description}
{data[0]}
```zathurarc
{data[1]}
```"""
    schemas[filetype]["properties"]["include"]["type"] = "array"
    schemas[filetype]["properties"]["include"]["uniqueItems"] = True
    for key in {"map", "unmap"}:
        schemas[filetype]["properties"][key]["type"] = "object"
        schemas[filetype]["properties"][key]["additionalProperties"] = False
        schemas[filetype]["properties"][key]["properties"] = {}
    # Use list not dict to keep order
    anyOf = [
        {
            "type": "string",
            "enum": keys,
        },
        {
            "type": "string",
            "pattern": "<[ACS]-.*>|.|[a-zA-Z]+",
        },
    ]
    for mode in ["normal", "fullscreen", "presentation", "index"]:
        schemas[filetype]["properties"]["unmap"]["properties"][mode] = {
            "type": "array",
            "uniqueItems": True,
            "items": {"anyOf": anyOf},
        }
        schemas[filetype]["properties"]["map"]["properties"][mode] = {
            "type": "array",
            "uniqueItems": True,
            "items": {
                "type": "object",
                "additionalProperties": False,
                "properties": {
                    "key": {"anyOf": anyOf},
                    "function": {"type": "string", "enum": functions},
                    "argument": {"type": "string", "enum": arguments + [None]},
                },
            },
        }

    schemas[filetype]["properties"]["set"]["properties"] = {}
    schemas[filetype]["properties"]["set"]["additionalProperties"] = False
    indices = []
    for i, token in enumerate(tokens[end_index:], end_index):
        if token.content == "SEE ALSO":
            end_index = i
            break
        if (
            token.type == "inline"
            and token.content.islower()
            and token.content.startswith("***")
            and token.content.endswith("***")
        ):
            indices += [i]
    for i, index in enumerate(indices):
        keyword = tokens[index].content.strip("*")
        index2 = end_index if len(indices) - 1 == i else indices[i + 1]
        for token in tokens[index + 1 : index2]:
            if (
                token.content != ""
                and not token.content.startswith("<!--")
                and token.content != ":"
            ):
                if (
                    schemas[filetype]["properties"]["set"]["properties"]
                    .get(keyword, {})
                    .get("description")
                    is None
                ):
                    description = " ".join(
                        line.strip()
                        for line in token.content.lstrip(": ").splitlines()
                    )
                    schemas[filetype]["properties"]["set"]["properties"][
                        keyword
                    ] = {"description": description}
                    if (
                        description.find("Possible values are ")
                        == description.find("Possible options are ")
                        == -1
                    ):
                        continue
                    schemas[filetype]["properties"]["set"]["properties"][
                        keyword
                    ]["enum"] = [
                        value.split(" ")[0].replace('\\"', "")
                        for value in description.rpartition(
                            "Possible values are "
                        )[2]
                        .rpartition("Possible options are ")[2]
                        .partition(".")[0]
                        .replace(" and ", ", ")
                        .split(", ")
                    ]
                    continue
                for line in token.content.splitlines():
                    if line.find("Value type: ") != -1:
                        _type = (
                            line.strip().partition("Value type: ")[2].lower()
                        )
                        schemas[filetype]["properties"]["set"]["properties"][
                            keyword
                        ]["type"] = {
                            "int": "integer",
                            "float": "number",
                            "bool": "boolean",
                        }.get(_type, _type)
                    elif line.find("Default value: ") != -1:
                        default = line.strip().partition("Default value: ")[2]
                        _type = schemas[filetype]["properties"]["set"][
                            "properties"
                        ][keyword]["type"]
                        if _type == "integer":
                            default = int(default.split(" ")[0])
                        elif _type == "number":
                            default = float(default)
                        elif _type == "boolean":
                            default = default == "true"
                        schemas[filetype]["properties"]["set"]["properties"][
                            keyword
                        ]["default"] = default
                        if isinstance(default, str) and default.startswith(
                            "#"
                        ):
                            schemas[filetype]["properties"]["set"][
                                "properties"
                            ][keyword]["format"] = "color"

    schemas[filetype]["properties"]["set"]["properties"][
        "highlight-transparency"
    ] |= {"minimum": 0, "maximum": 1}

    schemas[filetype]["properties"]["set"]["properties"]["guioptions"][
        "pattern"
    ] = r"[cshv]*"
    schemas[filetype]["properties"]["set"]["properties"]["first-page-column"][
        "pattern"
    ] = r"\d+(:\d+)*"
    schemas[filetype]["properties"]["set"]["properties"][
        "selection-clipboard"
    ]["enum"] = ["clipboard", "primary"]

    return schemas



r"""Test utils."""

import os

from zathura_language_server.finders import ImportZathurarcFinder
from zathura_language_server.utils import get_schema, parser


class Test:
    r"""Test."""

    @staticmethod
    def test_get_schema() -> None:
        r"""Test get schema.

        :rtype: None
        """
        assert len(
            get_schema()
            .get("properties", {})
            .get("set", {})
            .get("description", "")
            .splitlines()
        )

    @staticmethod
    def test_ImportZathurarcFinder() -> None:
        with open(
            os.path.join(os.path.dirname(__file__), "zathurarc"), "rb"
        ) as f:
            text = f.read()
        tree = parser.parse(text)
        assert ImportZathurarcFinder().get_document_links("", tree)



"""This file is generated by setuptools-generate.
The information comes from pyproject.toml.
It provide some metainfo for docs/conf.py to build documents and
help2man to build man pages.
"""

# For docs/conf.py
project = "zathura-language-server"
author = """Wu, Zhenyu <wuzhenyu@ustc.edu> """
copyright = "2025"

# For help2man
DESCRIPTION = "language server for zathurarc"
EPILOG = "Report bugs to https://github.com/Freed-Wu/zathura-language-server/issues"
# format __version__ by yourself
VERSION = """zathura-language-server {__version__}
Copyright (C) 2025
Written by Wu, Zhenyu <wuzhenyu@ustc.edu> """
SOURCE = "https://github.com/Freed-Wu/zathura-language-server"


r"""This module can be called by
`python -m <https://docs.python.org/3/library/__main__.html>`_.
"""

from argparse import ArgumentParser, RawDescriptionHelpFormatter
from datetime import datetime

from . import FILETYPE, __version__
from . import __name__ as NAME

try:
    import shtab
except ImportError:
    from . import _shtab as shtab

NAME = NAME.replace("_", "-")
VERSION = rf"""{NAME} {__version__}
Copyright (C) {datetime.now().year}
Written by Wu, Zhenyu
"""
EPILOG = """
Report bugs to <wuzhenyu@ustc.edu>.
"""


def get_parser():
    r"""Get a parser for unit test."""
    parser = ArgumentParser(
        epilog=EPILOG,
        formatter_class=RawDescriptionHelpFormatter,
    )
    shtab.add_argument_to(parser)
    parser.add_argument("--version", version=VERSION, action="version")
    parser.add_argument(
        "--generate-schema",
        choices=FILETYPE.__args__,  # type: ignore
        help="generate schema in an output format",
    )
    parser.add_argument(
        "--output-format",
        choices=["json", "yaml", "toml"],
        default="json",
        help="output format: %(default)s",
    )
    parser.add_argument(
        "--indent",
        type=int,
        default=2,
        help="generated json, yaml's indent, ignored by toml: %(default)s",
    )
    parser.add_argument(
        "--color",
        choices=["auto", "always", "never"],
        default="auto",
        help="when to display color, default: %(default)s",
    )
    parser.add_argument(
        "--check",
        nargs="*",
        default={},
        help="check file's errors and warnings",
    ).complete = shtab.FILE  # type: ignore
    parser.add_argument(
        "--convert",
        nargs="*",
        default={},
        help="convert files to output format",
    ).complete = shtab.FILE  # type: ignore
    return parser


def main() -> None:
    r"""Parse arguments and provide shell completions."""
    args = get_parser().parse_args()

    if args.generate_schema or args.check or args.convert:
        from lsp_tree_sitter.diagnose import check
        from lsp_tree_sitter.utils import pprint

        from .finders import DIAGNOSTICS_FINDER_CLASSES
        from .schema import ZathurarcTrie
        from .utils import parser

        if args.generate_schema:
            from .misc import get_schema

            kwargs = (
                {"indent": args.indent} if args.output_format != "toml" else {}
            )
            pprint(
                get_schema(args.generate_schema),
                filetype=args.output_format,
                **kwargs,
            )
        for file in args.convert:
            pprint(
                ZathurarcTrie.from_file(file, parser.parse).to_json(),
                filetype=args.output_format,
                indent=args.indent,
            )
        exit(
            check(
                args.check,
                parser.parse,
                DIAGNOSTICS_FINDER_CLASSES,
                None,
                args.color,
            )
        )

    from .server import ZathuraLanguageServer

    ZathuraLanguageServer(NAME, __version__).start_io()


if __name__ == "__main__":
    main()



r"""Utils
=========
"""

import json
import os
from typing import Any

from tree_sitter import Language, Parser, Query
from tree_sitter_zathurarc import language as get_language_ptr

from . import FILETYPE

SCHEMAS = {}
QUERIES = {}
parser = Parser()
parser.language = Language(get_language_ptr())


def get_query(name: str, filetype: FILETYPE = "zathurarc") -> Query:
    r"""Get query.

    :param name:
    :type name: str
    :param filetype:
    :type filetype: FILETYPE
    :rtype: Query
    """
    if name not in QUERIES:
        with open(
            os.path.join(
                os.path.dirname(__file__),
                "assets",
                "queries",
                f"{name}{os.path.extsep}scm",
            )
        ) as f:
            text = f.read()
        if parser.language:
            QUERIES[name] = Query(parser.language, text)
        else:
            raise NotImplementedError
    return QUERIES[name]


def get_schema(filetype: FILETYPE = "zathurarc") -> dict[str, Any]:
    r"""Get schema.

    :param filetype:
    :type filetype: FILETYPE
    :rtype: dict[str, Any]
    """
    if filetype not in SCHEMAS:
        file = os.path.join(
            os.path.dirname(__file__),
            "assets",
            "json",
            f"{filetype}.json",
        )
        with open(file) as f:
            SCHEMAS[filetype] = json.load(f)
    return SCHEMAS[filetype]
